# coding: utf-8
import ast
from typing import Iterable

from dataclasses import dataclass
from transpiler.phases.typing.scope import Scope
from transpiler.phases.emit_cpp import NodeVisitor, FunctionEmissionKind


class ClassVisitor(NodeVisitor):
    def visit_ClassDef(self, node: ast.ClassDef) -> Iterable[str]:
        yield "struct {"

        yield "struct type {"
        inner = ClassInnerVisitor(node.inner_scope)
        for stmt in node.body:
            yield from inner.visit(stmt)

        yield "template<typename... T> type(T&&... args) {"
        yield "__init__(std::forward<T>(args)...);"
        yield "}"
        yield "type(const type&) = delete;"
        yield "type(type&&) = delete;"
        yield "};"

        yield "template<typename... T> auto operator()(T&&... args) {"
        yield "return pyobj<type>(std::forward<T>(args)...);"
        yield "}"

        outer = ClassOuterVisitor(node.inner_scope)
        for stmt in node.body:
            yield from outer.visit(stmt)

        yield f"}} {node.name};"

@dataclass
class ClassInnerVisitor(NodeVisitor):
    scope: Scope

    def visit_AnnAssign(self, node: ast.AnnAssign) -> Iterable[str]:
        member = self.scope.obj_type.members[node.target.id]
        yield from self.visit(member)
        yield node.target.id
        yield ";"

    def visit_FunctionDef(self, node: ast.FunctionDef) -> Iterable[str]:
        yield "struct {"
        yield "type* self;"
        from transpiler.phases.emit_cpp.block import BlockVisitor
        yield from BlockVisitor(self.scope).visit_func_new(node, FunctionEmissionKind.METHOD, True)
        yield f"}} {node.name} {{ this }};"

@dataclass
class ClassOuterVisitor(NodeVisitor):
    scope: Scope

    def visit_AnnAssign(self, node: ast.AnnAssign) -> Iterable[str]:
        yield ""

    def visit_FunctionDef(self, node: ast.FunctionDef) -> Iterable[str]:
        yield "struct {"
        yield "template<typename... T>"
        yield "auto operator()(type& self, T&&... args) {"
        yield f"return self.{node.name}(std::forward<T>(args)...);"
        yield "}"
        yield f"}} {node.name};"
